import logging

import requests
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.http import HttpResponseRedirect, StreamingHttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema
from pulp_ansible.app.galaxy.v3 import views as pulp_ansible_views
from pulp_ansible.app.models import AnsibleDistribution
from pulp_ansible.app.models import CollectionImport as PulpCollectionImport
from pulp_ansible.app.models import CollectionVersion
from pulpcore.plugin.models import Task
from pulpcore.plugin.serializers import AsyncOperationResponseSerializer
from pulpcore.plugin.tasking import dispatch
from pulpcore.plugin.viewsets import OperationPostponedResponse
from rest_framework import status
from rest_framework.exceptions import APIException, NotFound
from rest_framework.request import Request
from rest_framework.response import Response
from semantic_version import SimpleSpec, Version

from galaxy_ng.app import models
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app.api import base as api_base
from galaxy_ng.app.api.v3.serializers import (
    CollectionSerializer,
    CollectionUploadSerializer,
    CollectionVersionListSerializer,
    CollectionVersionSerializer,
    UnpaginatedCollectionVersionSerializer,
)
from galaxy_ng.app.common import metrics
from galaxy_ng.app.common.parsers import AnsibleGalaxy29MultiPartParser
from galaxy_ng.app.constants import INBOUND_REPO_NAME_FORMAT, DeploymentMode
from galaxy_ng.app.tasks import (
    call_move_content_task,
    curate_all_synclist_repository,
    delete_collection,
    delete_collection_version,
    import_and_auto_approve,
    import_and_move_to_staging,
)

log = logging.getLogger(__name__)


class ViewNamespaceSerializerContextMixin:
    def get_serializer_context(self):
        """Inserts distribution path to a serializer context."""

        context = super().get_serializer_context()

        # view_namespace will be used by the serializers that need to return different hrefs
        # depending on where in the urlconf they are.
        # view_route is the url 'route' pattern, used to
        # handle the special case /api/automation-hub/v3/collections/ not having
        # a <str:path> in it's url
        request = context.get("request", None)
        context["view_namespace"] = None
        if request:
            context["view_namespace"] = request.resolver_match.namespace
            context["view_route"] = request.resolver_match.route

        return context


class RepoMetadataViewSet(api_base.LocalSettingsMixin,
                          pulp_ansible_views.RepoMetadataViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]


class UnpaginatedCollectionViewSet(api_base.LocalSettingsMixin,
                                   ViewNamespaceSerializerContextMixin,
                                   pulp_ansible_views.UnpaginatedCollectionViewSet):
    pagination_class = None
    permission_classes = [access_policy.CollectionAccessPolicy]
    serializer_class = CollectionSerializer


def get_collection_dependents(parent):
    """Given a parent collection, return a list of collection versions that depend on it."""
    key = f"{parent.namespace}.{parent.name}"
    dependents = []
    for child in CollectionVersion.objects.exclude(collection=parent).filter(
        dependencies__has_key=key
    ):
        dependents.append(child)
    return dependents


class CollectionViewSet(api_base.LocalSettingsMixin,
                        ViewNamespaceSerializerContextMixin,
                        pulp_ansible_views.CollectionViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]
    serializer_class = CollectionSerializer

    @extend_schema(
        description="Trigger an asynchronous delete task",
        responses={status.HTTP_202_ACCEPTED: AsyncOperationResponseSerializer},
    )
    def destroy(self, request: Request, *args, **kwargs) -> Response:
        """
        Allow a Collection to be deleted.
        1. Perform Dependency Check to verify that each CollectionVersion
           inside Collection can be deleted
        2. If the Collection can’t be deleted, return the reason why
        3. If it can, dispatch task to delete each CollectionVersion
           and the Collection
        """
        collection = self.get_object()

        # dependency check
        dependents = get_collection_dependents(collection)
        if dependents:
            return Response(
                {
                    "detail": _(
                        "Collection {namespace}.{name} could not be deleted "
                        "because there are other collections that require it."
                    ).format(
                        namespace=collection.namespace,
                        name=collection.name,
                    ),
                    "dependent_collection_versions": [
                        f"{dep.namespace}.{dep.name} {dep.version}" for dep in dependents
                    ],
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        repositories = set()
        for version in collection.versions.all():
            for repo in version.repositories.all():
                repositories.add(repo)

        async_result = dispatch(
            delete_collection,
            exclusive_resources=list(repositories),
            kwargs={"collection_pk": collection.pk},
        )

        return OperationPostponedResponse(async_result, request)


class UnpaginatedCollectionVersionViewSet(
    api_base.LocalSettingsMixin,
    ViewNamespaceSerializerContextMixin,
    pulp_ansible_views.UnpaginatedCollectionVersionViewSet,
):
    pagination_class = None
    serializer_class = UnpaginatedCollectionVersionSerializer
    permission_classes = [access_policy.CollectionAccessPolicy]


def get_dependents(parent):
    """Given a parent collection version, return a list of
    collection versions that depend on it.
    """
    key = f"{parent.namespace}.{parent.name}"
    dependents = []
    for child in CollectionVersion.objects.filter(dependencies__has_key=key):
        spec = SimpleSpec(child.dependencies[key])
        if spec.match(Version(parent.version)):
            dependents.append(child)
    return dependents


class CollectionVersionViewSet(api_base.LocalSettingsMixin,
                               ViewNamespaceSerializerContextMixin,
                               pulp_ansible_views.CollectionVersionViewSet):
    serializer_class = CollectionVersionSerializer
    permission_classes = [access_policy.CollectionAccessPolicy]
    list_serializer_class = CollectionVersionListSerializer

    @extend_schema(
        description="Trigger an asynchronous delete task",
        responses={status.HTTP_202_ACCEPTED: AsyncOperationResponseSerializer},
    )
    def destroy(self, request: Request, *args, **kwargs) -> Response:
        """
        Allow a CollectionVersion to be deleted.
        1. Perform Dependency Check to verify that the collection version can be deleted
        2. If the collection version can’t be deleted, return the reason why
        3. If it can, dispatch task to delete CollectionVersion and clean up repository.
           If the version being deleted is the last collection version in the collection,
           remove the collection object as well.
        """
        collection_version = self.get_object()

        # dependency check
        dependents = get_dependents(collection_version)
        if dependents:
            return Response(
                {
                    "detail": _(
                        "Collection version {namespace}.{name} {version} could not be "
                        "deleted because there are other collections that require it."
                    ).format(
                        namespace=collection_version.namespace,
                        name=collection_version.collection.name,
                        version=collection_version.version,
                    ),
                    "dependent_collection_versions": [
                        f"{dep.namespace}.{dep.name} {dep.version}" for dep in dependents
                    ],
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        async_result = dispatch(
            delete_collection_version,
            exclusive_resources=collection_version.repositories.all(),
            kwargs={"collection_version_pk": collection_version.pk},
        )

        return OperationPostponedResponse(async_result, request)


class CollectionVersionDocsViewSet(api_base.LocalSettingsMixin,
                                   pulp_ansible_views.CollectionVersionDocsViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]


class CollectionImportViewSet(api_base.LocalSettingsMixin,
                              pulp_ansible_views.CollectionImportViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]


class CollectionUploadViewSet(api_base.LocalSettingsMixin,
                              pulp_ansible_views.CollectionUploadViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]
    parser_classes = [AnsibleGalaxy29MultiPartParser]
    serializer_class = CollectionUploadSerializer

    def _dispatch_import_collection_task(self, temp_file_pk, repository=None, **kwargs):
        """Dispatch a pulp task started on upload of collection version."""
        locks = []
        context = super().get_serializer_context()
        request = context.get("request", None)

        kwargs["temp_file_pk"] = temp_file_pk
        kwargs["username"] = request.user.username

        if repository:
            locks.append(repository)
            kwargs["repository_pk"] = repository.pk

        if settings.GALAXY_REQUIRE_CONTENT_APPROVAL:
            return dispatch(import_and_move_to_staging, locks, kwargs=kwargs)
        return dispatch(import_and_auto_approve, locks, kwargs=kwargs)

    # Wrap super().create() so we can create a galaxy_ng.app.models.CollectionImport based on the
    # the import task and the collection artifact details

    def _get_data(self, request):
        serializer = CollectionUploadSerializer(data=request.data, context={'request': request})
        serializer.is_valid(raise_exception=True)

        return serializer.validated_data

    @staticmethod
    def _get_path(kwargs, filename_ns):
        """Use path from '/content/<path>/v3/' or
           if user does not specify distribution base path
           then use an inbound distribution based on filename namespace.
        """
        path = kwargs['path']
        if kwargs.get('no_path_specified', None):
            path = INBOUND_REPO_NAME_FORMAT.format(namespace_name=filename_ns)
        return path

    @staticmethod
    def _check_path_matches_expected_repo(path, filename_ns):
        """Reject if path does not match expected inbound format
           containing filename namespace.

        Examples:
        Reject if path is "staging".
        Reject if path does not start with "inbound-".
        Reject if path is "inbound-alice" but filename namepace is "bob".
        """

        distro = get_object_or_404(AnsibleDistribution, base_path=path)
        repo_name = distro.repository.name
        if INBOUND_REPO_NAME_FORMAT.format(namespace_name=filename_ns) == repo_name:
            return
        raise NotFound(
            _('Path does not match: "%s"')
            % INBOUND_REPO_NAME_FORMAT.format(namespace_name=filename_ns)
        )

    @extend_schema(
        description="Create an artifact and trigger an asynchronous task to create "
        "Collection content from it.",
        summary="Upload a collection",
        request=CollectionUploadSerializer,
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request, *args, **kwargs):
        data = self._get_data(request)
        filename = data['filename']

        path = self._get_path(kwargs, filename_ns=filename.namespace)

        try:
            namespace = models.Namespace.objects.get(name=filename.namespace)
        except models.Namespace.DoesNotExist:
            raise ValidationError(
                _('Namespace "{0}" does not exist.').format(filename.namespace)
            )

        self._check_path_matches_expected_repo(path, filename_ns=namespace.name)

        self.check_object_permissions(request, namespace)

        try:
            response = super(CollectionUploadViewSet, self).create(request, path)
        except ValidationError:
            log.exception('Failed to publish artifact %s (namespace=%s, sha256=%s)',  # noqa
                          data['file'].name, namespace, data.get('sha256'))
            raise

        task_href = response.data['task']

        # icky, have to extract task id from the task_href url
        task_id = task_href.strip("/").split("/")[-1]

        task_detail = Task.objects.get(pk=task_id)

        pulp_collection_import = PulpCollectionImport.objects.get(pk=task_id)

        models.CollectionImport.objects.create(
            task_id=pulp_collection_import,
            created_at=task_detail.pulp_created,
            namespace=namespace,
            name=data['filename'].name,
            version=data['filename'].version,
        )

        # TODO: CollectionImport.get_absolute_url() should be able to generate this, but
        #       it needs the  repo/distro base_path for the <path> part of url
        import_obj_url = reverse("galaxy:api:content:v3:collection-import",
                                 kwargs={'pk': str(task_detail.pk),
                                         'path': path})

        log.debug('import_obj_url: %s', import_obj_url)
        return Response(
            data={'task': import_obj_url},
            status=response.status_code
        )


class CollectionArtifactDownloadView(api_base.APIView):
    permission_classes = [access_policy.CollectionAccessPolicy]
    action = 'download'

    def _get_tcp_response(self, url):
        return requests.get(url, stream=True, allow_redirects=False)

    def _get_ansible_distribution(self, base_path):
        return AnsibleDistribution.objects.get(base_path=base_path)

    def get(self, request, *args, **kwargs):
        metrics.collection_artifact_download_attempts.inc()

        distro_base_path = self.kwargs['path']
        filename = self.kwargs['filename']
        prefix = settings.CONTENT_PATH_PREFIX.strip('/')
        distribution = self._get_ansible_distribution(self.kwargs['path'])

        if settings.GALAXY_DEPLOYMENT_MODE == DeploymentMode.INSIGHTS.value:
            url = 'http://{host}:{port}/{prefix}/{distro_base_path}/{filename}'.format(
                host=settings.X_PULP_CONTENT_HOST,
                port=settings.X_PULP_CONTENT_PORT,
                prefix=prefix,
                distro_base_path=distro_base_path,
                filename=filename,
            )
            response = self._get_tcp_response(url)
            response = redirect(distribution.content_guard.cast().preauthenticate_url(url))

            if response.status_code == requests.codes.not_found:
                metrics.collection_artifact_download_failures.labels(
                    status=requests.codes.not_found
                ).inc()
                raise NotFound()
            if response.status_code == requests.codes.found:
                return HttpResponseRedirect(response.headers['Location'])
            if response.status_code == requests.codes.ok:
                metrics.collection_artifact_download_successes.inc()
                return StreamingHttpResponse(
                    response.raw.stream(amt=4096),
                    content_type=response.headers['Content-Type']
                )
            metrics.collection_artifact_download_failures.labels(status=response.status_code).inc()
            raise APIException(
                _('Unexpected response from content app. Code: %s.') % response.status_code
            )
        elif settings.GALAXY_DEPLOYMENT_MODE == DeploymentMode.STANDALONE.value:
            url = '{host}/{prefix}/{distro_base_path}/{filename}'.format(
                host=settings.CONTENT_ORIGIN.strip("/"),
                prefix=prefix,
                distro_base_path=distro_base_path,
                filename=filename,
            )
            return redirect(distribution.content_guard.cast().preauthenticate_url(url))


class CollectionVersionMoveViewSet(api_base.ViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]

    def move_content(self, request, *args, **kwargs):
        """Remove content from source repo and add to destination repo.

        Creates new RepositoryVersion of source repo without content included.
        Creates new RepositoryVersion of destination repo with content included.
        """

        version_str = '-'.join([self.kwargs[key] for key in ('namespace', 'name', 'version')])
        try:
            collection_version = CollectionVersion.objects.get(
                namespace=self.kwargs['namespace'],
                name=self.kwargs['name'],
                version=self.kwargs['version'],
            )
        except ObjectDoesNotExist:
            raise NotFound(_('Collection %s not found') % version_str)

        try:
            src_repo = AnsibleDistribution.objects.get(
                base_path=self.kwargs['source_path']).repository
            dest_repo = AnsibleDistribution.objects.get(
                base_path=self.kwargs['dest_path']).repository
        except ObjectDoesNotExist:
            raise NotFound(_('Repo(s) for moving collection %s not found') % version_str)

        src_versions = CollectionVersion.objects.filter(pk__in=src_repo.latest_version().content)
        if collection_version not in src_versions:
            raise NotFound(_('Collection %s not found in source repo') % version_str)

        dest_versions = CollectionVersion.objects.filter(pk__in=dest_repo.latest_version().content)
        if collection_version in dest_versions:
            raise NotFound(_('Collection %s already found in destination repo') % version_str)

        move_task = call_move_content_task(collection_version, src_repo, dest_repo)

        curate_task_id = None
        if settings.GALAXY_DEPLOYMENT_MODE == DeploymentMode.INSIGHTS.value:
            golden_repo = AnsibleDistribution.objects.get(
                base_path=settings.GALAXY_API_DEFAULT_DISTRIBUTION_BASE_PATH
            ).repository

            if dest_repo == golden_repo or src_repo == golden_repo:
                repo_name = golden_repo.name
                locks = [golden_repo]
                task_args = (repo_name,)
                task_kwargs = {}

                curate_task = dispatch(
                    curate_all_synclist_repository, locks, args=task_args, kwargs=task_kwargs
                )
                curate_task_id = curate_task.pk

        return Response(
            data={
                "copy_task_id": move_task.pk,
                "remove_task_id": move_task.pk,
                "curate_all_synclist_repository_task_id": curate_task_id,
            },
            status='202'
        )



from django.conf import settings

from .base import BaseTestCase, get_current_ui_url


class TestUiFeatureFlagsView(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.original_setting = settings.GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_ACCESS
        settings.GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_ACCESS = True

    def tearDown(self):
        super().tearDown()
        settings.GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_ACCESS = self.original_setting

    def test_settings_url(self):
        self.settings_url = get_current_ui_url('settings')
        response = self.client.get(self.settings_url)
        self.assertEqual(response.data['GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_ACCESS'], True)
        self.assertEqual(response.data['GALAXY_ENABLE_UNAUTHENTICATED_COLLECTION_DOWNLOAD'], False)



import galaxy_importer

from django.apps import apps
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from pulp_ansible.app.models import AnsibleDistribution

from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.reverse import reverse

from galaxy_ng.app.api import base as api_base


# define the version matrix at the module level to avoid the redefinition on every API call.
VERSIONS = {
    "available_versions": {"v3": "v3/"},
    "server_version": apps.get_app_config("galaxy").version,
    "galaxy_ng_version": apps.get_app_config("galaxy").version,
    "galaxy_importer_version": galaxy_importer.__version__,
    "pulp_core_version": apps.get_app_config('core').version,
    "pulp_ansible_version": apps.get_app_config('ansible').version,
    "pulp_container_version": apps.get_app_config('container').version,
}


class ApiRootView(api_base.APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request, *args, **kwargs):
        """
        Returns the version matrix for the API + the current distro.
        """
        data = {**VERSIONS}

        if kwargs.get("path"):
            distro = get_object_or_404(
                AnsibleDistribution,
                base_path=self.kwargs["path"]
            )
            data["distro_base_path"] = distro.base_path

        return Response(data)


class ApiRedirectView(api_base.APIView):
    permission_classes = [IsAuthenticated]

    """Redirect requests to /api/automation-hub/api/ to /api/automation-hub/

    This is a workaround for https://github.com/ansible/ansible/issues/62073.
    This can be removed when ansible-galaxy stops appending '/api' to the url."""

    def get(self, request, *args, **kwargs):
        reverse_url_name = kwargs.get("reverse_url_name")

        reverse_kwargs = {}
        if "path" in kwargs:
            reverse_kwargs["path"] = kwargs["path"]

        return HttpResponseRedirect(reverse(reverse_url_name,
                                            kwargs=reverse_kwargs), status=307)



import logging

from django.conf import settings

log = logging.getLogger(__name__)
log.debug('openapi autoschema')


def preprocess_debug_logger(endpoints, **kwargs):
    """Log and then return an iterable of (path, path_regex, method, callback) tuples"""

    log.debug("kwargs: %s", repr(kwargs))

    for path, path_regex, method, callback in endpoints:
        log.debug('path=%s, path_regex=%s, method=%s, callback=%s',
                  path, path_regex, method, callback)

    return endpoints
    # return [
    #     (path, path_regex, method, callback) for path, path_regex, method, callback in endpoints
    #     if log.debug('path=%s, path_regex=%s, method=%s, callback=%s',
    #                  path, path_regex, method, callback)
    # ]


def preprocess_exclude_endpoints(endpoints, **kwargs):
    """Return an iterable of (path, path_regex, method, callback) with some endpoints removed

    For example, the default is to to remove '/pulp' and '/_ui/' api endpoints, while
    allowing /pulp/api/v3/status endpoint and all automation-hub v3 endpoints.
    """
    return [
        (path, path_regex, method, callback) for path, path_regex, method, callback in endpoints
        if path == '/pulp/api/v3/status/' or (not path.startswith('/pulp') and '/_ui/' not in path)
    ]


class AllowCorsMiddleware:
    """This is for use on dev environment and testing only"""

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        response = self.get_response(request)
        response["Access-Control-Allow-Origin"] = settings.get("GALAXY_CORS_ALLOWED_ORIGINS", "")
        response["Access-Control-Allow-Headers"] = settings.get("GALAXY_CORS_ALLOWED_HEADERS", "")

        # Code to be executed for each request/response after
        # the view is called.

        return response



import yaml
from django.core.exceptions import ObjectDoesNotExist
from galaxy_ng.app import models
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app.api import base as api_base
from pulp_ansible.app.models import AnsibleDistribution
from rest_framework.renderers import (
    BaseRenderer,
    BrowsableAPIRenderer,
    JSONRenderer
)
from rest_framework.request import Request
from rest_framework.response import Response
from yaml.dumper import SafeDumper


def get_synclist_excludes(base_path):
    """Get Synclist from distro base_path"""
    try:
        repo = AnsibleDistribution.objects.get(base_path=base_path).repository
        synclist = models.SyncList.objects.get(repository=repo, policy="exclude")
        return synclist.collections.all()
    except ObjectDoesNotExist:
        return None


def serialize_collection_queryset(queryset):
    """Serialize a Queryset in to a JSONable format."""
    return queryset is not None and [
        {
            "name": "{collection.namespace}.{collection.name}".format(collection=collection)
        }
        for collection in queryset.all()
    ] or []


class RequirementsFileRenderer(BaseRenderer):
    """Renders requirements YAML format."""

    media_type = 'application/yaml'
    format = 'yaml'
    charset = 'utf-8'
    encoder = SafeDumper

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Renders data to YAML.

        Args:
            data: Object to be rendered as YAML.
            accepted_media_type: The media type the client accepts.
            renderer_context: Renderer context.

        Returns:
            A rendered response in YAML format.
        """
        if data is None:
            return ""

        return yaml.dump(
            data,
            stream=None,
            encoding=self.charset,
            Dumper=self.encoder,
            allow_unicode=True,
            default_flow_style=False,
        )


class ExcludesView(api_base.APIView):
    permission_classes = [access_policy.CollectionAccessPolicy]
    action = 'list'
    renderer_classes = [
        JSONRenderer,
        BrowsableAPIRenderer,
        RequirementsFileRenderer
    ]

    def get(self, request: Request, *args, **kwargs):
        """
        Returns a list of excludes for a given distro.
        """
        queryset = get_synclist_excludes(self.kwargs["path"])
        collections_to_exclude = serialize_collection_queryset(queryset)
        return Response({"collections": collections_to_exclude})



# Generated by Django 2.2.16 on 2020-11-02 14:48

from django.db import migrations


class Migration(migrations.Migration):

    replaces = [('galaxy', '0012_move_collections_by_certification')]

    dependencies = [
        ('galaxy', '0011_collection_import_task_id_fk'),
    ]

    operations = []



from galaxy_ng.app.models.auth import Group

from django.contrib.auth import get_user_model
from django.core.management import BaseCommand

User = get_user_model()


class Command(BaseCommand):
    """
    Django management command for creating groups
    """

    help = 'Create an access group, and optionally assign it to one or more users'

    def add_arguments(self, parser):
        parser.add_argument('groups', nargs='+')
        parser.add_argument(
            '--users',
            default=[],
            nargs='*',
            help='Assign the group(s) to one or more users'
        )

    def handle(self, *args, **options):
        for group_name in options['groups']:
            group, created = Group.objects.get_or_create(name=group_name)
            if created:
                self.stdout.write("Created group '{}'".format(group_name))
            else:
                self.stdout.write("Group '{}' already exists".format(group_name))

            for username in options['users']:
                try:
                    user = User.objects.get(username=username)
                except User.DoesNotExist:
                    self.stdout.write("User '{}' not found. Skipping.".format(username))
                    continue
                user.groups.add(group)
                user.save()
                self.stdout.write("Assigned group '{}' to user '{}'".format(group_name, username))



import logging

from django.conf import settings
from django.apps import apps

from rest_framework import status as http_code

from galaxy_ng.app.models import auth as auth_models
from galaxy_ng.app.constants import DeploymentMode
from . import base

from .synclist_base import BaseSyncListViewSet, ACCOUNT_SCOPE

log = logging.getLogger(__name__)
logging.getLogger().setLevel(logging.DEBUG)

log.info("settings.FIXTURE_DIRS(module scope): %s", settings.FIXTURE_DIRS)


class TestUiMySyncListViewSet(BaseSyncListViewSet):
    def setUp(self):
        super().setUp()

        log.info("self.fixtures2: %s", self.fixtures)
        log.info("settings.FIXTURE_DIRS2: %s", settings.FIXTURE_DIRS)

        self.user = auth_models.User.objects.create_user(username="test1", password="test1-secret")
        self.group = self._create_group_with_synclist_perms(
            ACCOUNT_SCOPE, "test1_group", users=[self.user]
        )
        self.user.save()
        self.group.save()

        self.user.groups.add(self.group)
        self.user.save()

        self.synclist_name = "test_synclist"
        self.synclist = self._create_synclist(
            name=self.synclist_name,
            repository=self.repo,
            upstream_repository=self.default_repo,
            groups=[self.group],
        )

        self.client.force_authenticate(user=self.user)

    def test_my_synclist_create(self):
        with self.settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.INSIGHTS.value):
            galaxy_app_config = apps.get_app_config("galaxy")
            log.info("gac module: %s", galaxy_app_config.module)
            log.info("gac path: %s", galaxy_app_config.path)

            post_data = {
                "repository": self.repo.pk,
                "collections": [],
                "namespaces": [],
                "policy": "include",
                "groups": [
                    {
                        "id": self.group.id,
                        "name": self.group.name,
                        "object_permissions": self.default_owner_permissions,
                    },
                ],
            }

            synclists_url = base.get_current_ui_url("my-synclists-list")

            response = self.client.post(synclists_url, post_data, format="json")

            log.debug("response: %s", response)
            log.debug("response.data: %s", response.data)

            # synclist create is not allowed via my-synclist viewset
            self.assertEqual(response.status_code, http_code.HTTP_403_FORBIDDEN, msg=response.data)

    def test_my_synclist_update(self):
        with self.settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.INSIGHTS.value):
            ns_name = "unittestnamespace1"
            ns = self._create_namespace(ns_name, groups=[self.group])
            ns.save()

            post_data = {
                "repository": self.repo.pk,
                "collections": [],
                "namespaces": [ns_name],
                "policy": "include",
                "groups": [
                    {
                        "id": self.group.id,
                        "name": self.group.name,
                        "object_permissions": self.default_owner_permissions,
                    },
                ],
            }

            synclists_detail_url = base.get_current_ui_url(
                "my-synclists-detail", kwargs={"pk": self.synclist.id}
            )

            response = self.client.patch(synclists_detail_url, post_data, format="json")

            log.debug("response: %s", response)
            log.debug("response.data: %s", response.data)

            self.assertEqual(response.status_code, http_code.HTTP_200_OK)
            self.assertIn("name", response.data)
            self.assertIn("repository", response.data)
            self.assertEqual(response.data["name"], self.synclist_name)
            self.assertEqual(response.data["policy"], "include")

            # Sort permission list for comparison
            response.data["groups"][0]["object_permissions"].sort()
            self.default_owner_permissions.sort()
            self.assertEqual(
                response.data["groups"],
                [
                    {
                        "name": self.group.name,
                        "id": self.group.id,
                        "object_permissions": self.default_owner_permissions,
                    }
                ],
            )

    def test_my_synclist_list(self):
        with self.settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.INSIGHTS.value):
            synclists_url = base.get_current_ui_url("my-synclists-list")
            log.debug("synclists_url: %s", synclists_url)

            response = self.client.get(synclists_url)

            log.debug("response: %s", response)
            data = response.data["data"]

            self.assertIsInstance(data, list)
            self.assertEqual(len(data), 1)
            self.assertEqual(data[0]["name"], self.synclist_name)
            self.assertEqual(data[0]["policy"], "exclude")
            self.assertEqual(data[0]["repository"], self.repo.pk)

    def test_my_synclist_detail(self):
        with self.settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.INSIGHTS.value):
            synclists_detail_url = base.get_current_ui_url(
                "my-synclists-detail", kwargs={"pk": self.synclist.id}
            )

            log.debug("synclists_detail_url: %s", synclists_detail_url)

            response = self.client.get(synclists_detail_url)

            log.debug("response: %s", response)

            self.assertEqual(response.status_code, http_code.HTTP_200_OK)
            self.assertIn("name", response.data)
            self.assertIn("repository", response.data)
            self.assertEqual(response.data["name"], self.synclist_name)
            self.assertEqual(response.data["policy"], "exclude")
            self.assertEqual(response.data["collections"], [])
            self.assertEqual(response.data["namespaces"], [])

    def test_my_synclist_delete(self):
        with self.settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.INSIGHTS.value):
            synclists_detail_url = base.get_current_ui_url(
                "my-synclists-detail", kwargs={"pk": self.synclist.id}
            )

            log.debug("delete url: %s", synclists_detail_url)

            response = self.client.delete(synclists_detail_url)

            log.debug("delete response: %s", response)

            self.assertEqual(response.status_code, http_code.HTTP_403_FORBIDDEN)



# Generated by Django 3.2.6 on 2021-09-23 16:11

from django.db import migrations, models
import django.db.models.deletion
import galaxy_ng.app.models.contentguard


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0075_rbaccontentguard'),
        ('galaxy', '0023_auto_20210903_1906'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContentRedirectContentGuard',
            fields=[
                ('contentguard_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='galaxy_contentredirectcontentguard', serialize=False, to='core.contentguard')),
                ('shared_secret', models.BinaryField(default=galaxy_ng.app.models.contentguard._gen_secret, max_length=32)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.contentguard',),
        ),
    ]



from django.contrib.auth.models import Permission
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from guardian.shortcuts import get_perms_for_model

from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from galaxy_ng.app.models import auth as auth_models


class GroupPermissionField(serializers.Field):
    def _validate_group(self, group_data):
        if 'object_permissions' not in group_data:
            raise ValidationError(detail={
                'groups': _('object_permissions field is required')})

        if 'id' not in group_data and 'name' not in group_data:
            raise ValidationError(detail={
                'groups': _('id or name field is required')})

        perms = group_data['object_permissions']

        if not isinstance(perms, list):
            raise ValidationError(detail={
                'groups': _('object_permissions must be a list of strings')})

        # validate that the permissions exist
        for perm in perms:
            if '.' in perm:
                app_label, codename = perm.split('.', maxsplit=1)
                filter_q = Q(content_type__app_label=app_label) & Q(codename=codename)
            else:
                filter_q = Q(codename=perm)

            # TODO(newswangerd): Figure out how to make this one SQL query instead of
            # performing N queries for each permission
            if not Permission.objects.filter(filter_q).exists():
                raise ValidationError(detail={
                    'groups': _('Permission {} does not exist').format(perm)})

    def to_representation(self, value):
        rep = []
        for group in value:
            rep.append({
                'id': group.id,
                'name': group.name,
                'object_permissions': value[group]
            })
        return rep

    def to_internal_value(self, data):
        if not isinstance(data, list):
            raise ValidationError(detail={
                'groups': _('Groups must be a list of group objects')
            })

        internal = {}
        for group_data in data:
            self._validate_group(group_data)
            group_filter = {}
            for field in group_data:
                if field in ('id', 'name'):
                    group_filter[field] = group_data[field]
            try:
                group = auth_models.Group.objects.get(**group_filter)
                internal[group] = group_data['object_permissions']
            except auth_models.Group.DoesNotExist:
                raise ValidationError(detail={
                    'groups': _("Group name=%s, id=%s does not exist") % (
                        group_data.get('name'), group_data.get('id'))
                })
            except ValueError:
                raise ValidationError(detail={'group': _('Invalid group name or ID')})

        return internal


class MyPermissionsField(serializers.Serializer):
    def to_representation(self, obj):
        user = self.context['request'].user

        # guardian's get_perms(user, obj) method only returns user permissions,
        # not all permissions a user has.
        my_perms = []
        for perm in get_perms_for_model(type(obj)).all():
            codename = "{}.{}".format(perm.content_type.app_label, perm.codename)
            if user.has_perm(codename) or user.has_perm(codename, obj):
                my_perms.append(codename)

        return my_perms



from django.conf import settings
from rest_framework.permissions import IsAuthenticated

from galaxy_ng.app.api import base as api_base


class ControllerListView(api_base.APIView):
    permission_classes = [IsAuthenticated]

    # Returns a paginated list. This will make this easier to upgrade to a
    # database setting down the line.
    def get(self, request, *args, **kwargs):
        host_filter = request.GET.get("host", None)
        host_icontains_filter = request.GET.get("host__icontains", None)

        controllers = []
        for controller in settings.CONNECTED_ANSIBLE_CONTROLLERS:
            if host_filter and controller != host_filter:
                continue

            if host_icontains_filter and host_icontains_filter.lower() not in controller.lower():
                continue

            controllers.append({"host": controller})
        paginator = self.pagination_class()
        page = paginator.paginate_queryset(controllers, request, view=self)

        return paginator.get_paginated_response(page)



import unittest

from urllib.parse import urlparse

from pulpcore.client.galaxy_ng.exceptions import ApiException
from pulp_container.tests.functional.api import rbac_base
from pulp_container.tests.functional.constants import DOCKERHUB_PULP_FIXTURE_1
from pulp_smash import cli

from galaxy_ng.tests.functional.utils import ContainerTestCaseUsingBindings


class ContainerRepositoryTagsTestCase(ContainerTestCaseUsingBindings, rbac_base.BaseRegistryTest):
    """Test whether a container repository's tags can be listed.

    When running functional tests in dev environment please ensure that
    Pulp Smash can either execute commands as root or can successfully
    execute "sudo" on the localhost.
    .. note:: When running against a non-https registry the client config
        `insecure-registries` must be enabled.
    For docker it is located in `/etc/docker/daemon.json` and content is::
        {"insecure-registries": ["0.0.0.0:5001"]}
    For podman it is located in `/etc/containers/registries.conf` with::
        [registries.insecure]
        registries = ['0.0.0.0:5001']
    """

    @classmethod
    def setUpClass(cls):
        """
        Define APIs to use and pull images needed later in tests
        """
        super().setUpClass()

        cls.registry = cli.RegistryClient(cls.cfg)
        cls.registry.raise_if_unsupported(unittest.SkipTest, "Tests require podman/docker")
        cls.registry_name = urlparse(cls.cfg.get_base_url()).netloc
        admin_user, admin_password = cls.cfg.pulp_auth
        cls.user_admin = {"username": admin_user, "password": admin_password}
        cls._pull(f"{DOCKERHUB_PULP_FIXTURE_1}:manifest_a")

    def test_list_container_repository_tags(self):
        image_name = "foo/bar"
        image_path = f"{DOCKERHUB_PULP_FIXTURE_1}:manifest_a"
        local_url = "/".join([self.registry_name, f"{image_name}:1.0"])

        # expect the Container Repository Tags not to exist yet
        with self.assertRaises(ApiException) as context:
            self.container_repo_tags_api.list(base_path=image_name)

        # Bad request
        assert context.exception.status == 404

        self._push(image_path, local_url, self.user_admin)

        response = self.container_repo_tags_api.list(base_path=image_name)

        self.assertEqual(response.meta.count, 1)
        for entry in response.data:
            self.assertIn(entry.name, ["1.0"])

        # Delete created Execution Environment
        # api does not currently support delete
        ee_delete_response = self.smash_client.delete(
            f"{self.galaxy_api_prefix}/_ui/v1/execution-environments/repositories/{image_name}/"
        )
        print(f"Delete execution environment: {ee_delete_response['state']}")



from .standalone import STANDALONE_STATEMENTS
from .insights import INSIGHTS_STATEMENTS
from .pulp_container import PULP_CONTAINER_VIEWSETS

__all__ = (
    STANDALONE_STATEMENTS,
    INSIGHTS_STATEMENTS,
    PULP_CONTAINER_VIEWSETS
)



from pulp_ansible.app.models import Tag
from pulp_ansible.app.serializers import TagSerializer

from galaxy_ng.app.api import base as api_base
from galaxy_ng.app.access_control import access_policy

from galaxy_ng.app.api.ui import versioning


class TagsViewSet(api_base.GenericViewSet):
    serializer_class = TagSerializer
    permission_classes = [access_policy.TagsAccessPolicy]
    versioning_class = versioning.UIVersioning

    queryset = Tag.objects.all()

    def list(self, request, *args, **kwargs):
        qs = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(qs)
        serializer = self.get_serializer(page, many=True)

        return self.get_paginated_response(serializer.data)
